import pandas as pd


class ItypeWarning(RuntimeWarning):
    pass


class ItypeCastWarning(ItypeWarning):
    pass


class ItypeCastError(RuntimeError):
    pass


class __Itype:
    def __init__(self):
        raise RuntimeError("DatetimeItype does not allow instances of itself.")


class DatetimeItype(__Itype):
    name = 'datetime'
    unique = True
    subtypes = (pd.DatetimeIndex,)


class IntegerItype(__Itype):
    name = 'integer'
    unique = True
    subtypes = (pd.RangeIndex, pd.Int64Index, pd.UInt64Index, int)


class FloatItype(__Itype):
    name = 'float'
    subtypes = (pd.Float64Index, float)
    unique = True


# class MultiItype(__Itype):
#     name = "multi"
#     subtypes = (pd.MultiIndex, )
#     unique = ??


class NumericItype(__Itype):
    name = "numeric"
    _subitypes = (IntegerItype, FloatItype)
    subtypes = (_subitypes + IntegerItype.subtypes + FloatItype.subtypes)
    unique = False


class MixedItype(__Itype):
    name = "mixed"
    unique = False
    _subitypes = (DatetimeItype, IntegerItype, FloatItype, NumericItype)
    _otheritypes = (pd.CategoricalIndex, pd.IntervalIndex, pd.PeriodIndex, pd.TimedeltaIndex)
    subtypes = (_subitypes + _otheritypes + DatetimeItype.subtypes + NumericItype.subtypes)


def is_itype(obj, itype):
    """ Check if obj is a instance of the given itype or its str-alias was given"""

    # todo: iter through itype as it could be a tuple, if called like ``is_itype(o, (t1,t2))``

    # user gave a Itype, like ``DatetimeItype``
    if type(obj) == type and issubclass(obj, itype):
        return True

    # user gave a string, like 'datetime'
    if isinstance(obj, str) and obj == itype.name:
        return True

    return False


def is_itype_subtype(obj, itype):
    """ Check if obj is a subclass or a instance of a subclass of the given itype"""

    # user gave a subtype, like ``pd.DatetimeIndex``
    if type(obj) == type and issubclass(obj, itype.subtypes):
        return True

    # user gave a instance of a subtype, like ``pd.Series(..).index``
    if isinstance(obj, itype.subtypes):
        return True

    return False


def is_itype_like(obj, itype):
    """ Check if obj is a subclass or a instance of the given itype or any of its subtypes"""
    return is_itype(obj, itype) or is_itype_subtype(obj, itype)


def get_itype(obj):
    """
    Return the according Itype, by any of any possible user input, like
      - "datetime"
      - DatetimeItype
      - pd.Series(...).index
      - pd.DatetimeIndex
    and return the according Itype
    """
    if type(obj) == type and issubclass(obj, __Itype):
        return obj

    # check if it is the actual type, not a subtype
    types = [DatetimeItype, IntegerItype, FloatItype, NumericItype, MixedItype]
    for t in types:
        if is_itype(obj, t):
            return t

    # If the above failed, we try to infer the itype by its subtypes.
    # We just check the unique types, because the non-unique are just
    # collections of unique subtypes, and would have be detected by any
    # of the upper if-statements
    for t in types:
        if is_itype_subtype(obj, t) and t.unique:
            return t

    raise ValueError(f"{obj} is not a itype, nor any known subtype of a itype, nor a itype string alias")


def itype_eq(a, b):
    return is_itype(a, b)


def itype_lt(a, b):
    return is_itype_subtype(a, b)


def itype_le(a, b):
    return is_itype_like(a, b)


def cast_to_fit_itype(series, itype, policy='force'):
    """ Cast a series (more explicit the type of the index) to fit the itype of a dios.

    Return the casted series if successful, None otherwise.

    Note:
        This is very basic number-casting, so in most cases, information from
        the old index will be lost after the cast.
    """
    series.itype = get_itype(series.index)

    # up-cast issn't necessary because a dios with a higher
    # itype always can take lower itypes.
    # series can have dt/int/float/mixed
    # dt    -> dt           -> mixed
    # int   -> int   -> num -> mixed
    # float -> float -> num -> mixed
    # mixed                 -> mixed
    if itype_le(series.itype, itype):  # a <= b
        return series

    if policy in ['forbid', 'no-downcast', 'no-cast', 'never']:
        return None

    elif policy == 'force':
        # any (dt/float/mixed)      -> int,      always OK
        # any (dt/float/mixed)      -> num(int), always OK
        # any (dt/int/mixed)        -> float,    always OK
        # any (int/float/mixed)     -> dt,       always FAIL
        if is_itype(itype, IntegerItype) or is_itype(itype, NumericItype):  # a == b or a == c
            series.index = pd.RangeIndex(len(series))
            return series
        if is_itype(itype, FloatItype):  # a == b
            series.index = pd.Float64Index(range(len(series)))
            return series
        if is_itype(itype, DatetimeItype):  # a == b
            return None
        return None

    elif policy == 'lossless':
        # int   -> float, always OK
        # float -> int,   maybe if unique
        # mixed -> any,   always FAIL
        # dt    -> any,   always FAIL
        if is_itype(itype, IntegerItype) and is_itype(series.itype, FloatItype):  # a == b and c == d
            series.index = series.index.astype(float)
            return series
        if is_itype(itype, FloatItype) and is_itype(series.itype, IntegerItype):  # a == b and c == d
            series.index = series.index.astype(int)
            if series.index.is_unique:
                return series
        return None

    raise ValueError(f"Unknown policy `{policy}`.")

